@@ -25,18 +25,18 @@ import sys
 import html
 from typing import TYPE_CHECKING, Optional, Set
 
-from PyQt5.QtCore import QObject
-import PyQt5.QtCore as QtCore
-from PyQt5.QtWidgets import (QWidget, QLabel, QPushButton, QTextEdit,
-                             QMessageBox, QHBoxLayout, QVBoxLayout)
+from PyQt6.QtCore import QObject, Qt
+import PyQt6.QtCore as QtCore
+from PyQt6.QtWidgets import (QWidget, QLabel, QPushButton, QTextEdit,
+                             QMessageBox, QHBoxLayout, QVBoxLayout, QDialog, QScrollArea)
 
 from electrum.i18n import _
-from electrum.base_crash_reporter import BaseCrashReporter, EarlyExceptionsQueue
+from electrum.base_crash_reporter import BaseCrashReporter, EarlyExceptionsQueue, CrashReportResponse
 from electrum.logging import Logger
 from electrum import constants
 from electrum.network import Network
 
-from .util import MessageBoxMixin, read_QIcon, WaitingDialog
+from .util import MessageBoxMixin, read_QIcon, WaitingDialog, font_height
 
 if TYPE_CHECKING:
     from electrum.simple_config import SimpleConfig
@@ -65,18 +65,16 @@ class Exception_Window(BaseCrashReporter, QWidget, MessageBoxMixin, Logger):
 
         main_box.addWidget(QLabel(BaseCrashReporter.REQUEST_HELP_MESSAGE))
 
+        self._report_contents_dlg = None  # type: Optional[ReportContentsDialog]
         collapse_info = QPushButton(_("Show report contents"))
-        collapse_info.clicked.connect(
-            lambda: self.msg_box(QMessageBox.NoIcon,
-                                 self, _("Report contents"), self.get_report_string(),
-                                 rich_text=True))
+        collapse_info.clicked.connect(lambda _checked: self.show_report_contents_dlg())
 
         main_box.addWidget(collapse_info)
 
         main_box.addWidget(QLabel(BaseCrashReporter.DESCRIBE_ERROR_MESSAGE))
 
         self.description_textfield = QTextEdit()
-        self.description_textfield.setFixedHeight(50)
+        self.description_textfield.setFixedHeight(4 * font_height())
         self.description_textfield.setPlaceholderText(self.USER_COMMENT_PLACEHOLDER)
         main_box.addWidget(self.description_textfield)
 
@@ -85,32 +83,37 @@ class Exception_Window(BaseCrashReporter, QWidget, MessageBoxMixin, Logger):
         buttons = QHBoxLayout()
 
         report_button = QPushButton(_('Send Bug Report'))
-        report_button.clicked.connect(self.send_report)
+        report_button.clicked.connect(lambda _checked: self.send_report())
         report_button.setIcon(read_QIcon("tab_send.png"))
         buttons.addWidget(report_button)
 
         never_button = QPushButton(_('Never'))
-        never_button.clicked.connect(self.show_never)
+        never_button.clicked.connect(lambda _checked: self.show_never())
         buttons.addWidget(never_button)
 
         close_button = QPushButton(_('Not Now'))
-        close_button.clicked.connect(self.close)
+        close_button.clicked.connect(lambda _checked: self.close())
         buttons.addWidget(close_button)
 
         main_box.addLayout(buttons)
 
+        # prioritizes the window input over all other windows
+        self.setWindowModality(QtCore.Qt.WindowModality.ApplicationModal)
+
         self.setLayout(main_box)
         self.show()
 
     def send_report(self):
-        def on_success(response):
-            # note: 'response' coming from (remote) crash reporter server.
-            # It contains a URL to the GitHub issue, so we allow rich text.
+        def on_success(response: CrashReportResponse):
+            text = response.text
+            if response.url:
+                text += f" You can track further progress on <a href='{response.url}'>GitHub</a>."
             self.show_message(parent=self,
                               title=_("Crash report"),
-                              msg=response,
+                              msg=text,
                               rich_text=True)
             self.close()
+
         def on_failure(exc_info):
             e = exc_info[1]
             self.logger.error('There was a problem with the automatic reporting', exc_info=exc_info)
@@ -131,7 +134,7 @@ class Exception_Window(BaseCrashReporter, QWidget, MessageBoxMixin, Logger):
         self.close()
 
     def show_never(self):
-        self.config.set_key(BaseCrashReporter.config_key, False)
+        self.config.SHOW_CRASH_REPORTER = False
         self.close()
 
     def closeEvent(self, event):
@@ -152,6 +155,15 @@ class Exception_Window(BaseCrashReporter, QWidget, MessageBoxMixin, Logger):
         traceback_str = super()._get_traceback_str_to_display()
         return html.escape(traceback_str)
 
+    def show_report_contents_dlg(self):
+        if self._report_contents_dlg is None:
+            self._report_contents_dlg = ReportContentsDialog(
+                parent=self,
+                text=self.get_report_string(),
+            )
+        self._report_contents_dlg.show()
+        self._report_contents_dlg.raise_()
+
 
 def _show_window(*args):
     if not Exception_Window._active_window:
@@ -176,7 +188,7 @@ class Exception_Hook(QObject, Logger):
 
     @classmethod
     def maybe_setup(cls, *, config: 'SimpleConfig', wallet: 'Abstract_Wallet' = None) -> None:
-        if not config.get(BaseCrashReporter.config_key, default=True):
+        if not config.SHOW_CRASH_REPORTER:
             EarlyExceptionsQueue.set_hook_as_ready()  # flush already queued exceptions
             return
         if not cls._INSTANCE:
@@ -187,3 +199,20 @@ class Exception_Hook(QObject, Logger):
     def handler(self, *exc_info):
         self.logger.error('exception caught by crash reporter', exc_info=exc_info)
         self._report_exception.emit(self.config, *exc_info)
+
+
+class ReportContentsDialog(QDialog):
+
+    def __init__(self, *, parent: QWidget, text: str):
+        QDialog.__init__(self, parent)
+        self.setWindowTitle(_("Report contents"))
+        self.setMinimumSize(800, 500)
+        vbox = QVBoxLayout(self)
+        scroll_area = QScrollArea(self)
+
+        report_text = QLabel(text)
+        report_text.setTextInteractionFlags(Qt.TextInteractionFlag.TextSelectableByMouse)
+        report_text.setTextFormat(Qt.TextFormat.AutoText)  # likely rich text
+
+        scroll_area.setWidget(report_text)
+        vbox.addWidget(scroll_area)
