@@ -22,55 +22,53 @@
 # SOFTWARE.
 import asyncio
 import time
-import queue
 import os
 import random
 import re
 from collections import defaultdict
 import threading
-import socket
 import json
-import sys
-from typing import NamedTuple, Optional, Sequence, List, Dict, Tuple, TYPE_CHECKING, Iterable, Set, Any, TypeVar
-import traceback
-import concurrent
-from concurrent import futures
+from typing import (
+    NamedTuple, Optional, Sequence, List, Dict, Tuple, TYPE_CHECKING, Iterable, Set, Any, TypeVar,
+    Callable
+)
 import copy
 import functools
 from enum import IntEnum
+from contextlib import nullcontext
 
 import aiorpcx
-from aiorpcx import ignore_after
+from aiorpcx import ignore_after, NetAddress
 from aiohttp import ClientResponse
 
 from . import util
-from .util import (log_exceptions, ignore_exceptions, OldTaskGroup,
-                   bfh, make_aiohttp_session, send_exception_to_crash_reporter,
-                   is_hash256_str, is_non_negative_integer, MyEncoder, NetworkRetryManager,
-                   nullcontext, error_text_str_to_safe_str)
-from .bitcoin import COIN
+from .util import (
+    log_exceptions, ignore_exceptions, OldTaskGroup, make_aiohttp_session, send_exception_to_crash_reporter, MyEncoder,
+    NetworkRetryManager, error_text_str_to_safe_str, detect_tor_socks_proxy
+)
+from .bitcoin import DummyAddress, DummyAddressUsedInTxException
 from . import constants
 from . import blockchain
-from . import bitcoin
 from . import dns_hacks
 from .transaction import Transaction
-from .blockchain import Blockchain, HEADER_SIZE
-from .interface import (Interface, PREFERRED_NETWORK_PROTOCOL,
-                        RequestTimedOut, NetworkTimeout, BUCKET_NAME_OF_ONION_SERVERS,
-                        NetworkException, RequestCorrupted, ServerAddr)
+from .blockchain import Blockchain
+from .interface import (
+    Interface, PREFERRED_NETWORK_PROTOCOL, RequestTimedOut, NetworkTimeout, BUCKET_NAME_OF_ONION_SERVERS,
+    NetworkException, RequestCorrupted, ServerAddr
+)
 from .version import PROTOCOL_VERSION
-from .simple_config import SimpleConfig
 from .i18n import _
 from .logging import get_logger, Logger
+from .fee_policy import FeeHistogram, FeeTimeEstimates, FEE_ETA_TARGETS
+
 
 if TYPE_CHECKING:
     from collections.abc import Coroutine
-
     from .channel_db import ChannelDB
     from .lnrouter import LNPathFinder
     from .lnworker import LNGossip
-    from .lnwatcher import WatchTower
     from .daemon import Daemon
+    from .simple_config import SimpleConfig
 
 
 _logger = get_logger(__name__)
@@ -155,51 +153,137 @@ def pick_random_server(hostmap=None, *, allowed_protocols: Iterable[str],
     return random.choice(eligible) if eligible else None
 
 
+def is_valid_port(ps: str):
+    try:
+        return 0 < int(ps) < 65535
+    except ValueError:
+        return False
+
+
+def is_valid_host(ph: str):
+    try:
+        NetAddress(ph, '1')
+    except ValueError:
+        return False
+    return True
+
+
+class ProxySettings:
+    MODES = ['socks4', 'socks5']
+
+    probe_fut = None
+
+    def __init__(self):
+        self.enabled = False
+        self.mode = 'socks5'
+        self.host = ''
+        self.port = ''
+        self.user = None
+        self.password = None
+
+    def set_defaults(self):
+        self.__init__()  # call __init__ for default values
+
+    def serialize_proxy_cfgstr(self):
+        return ':'.join([self.mode, self.host, self.port])
+
+    def deserialize_proxy_cfgstr(self, s: Optional[str], user: str = None, password: str = None) -> None:
+        if s is None or (isinstance(s, str) and s.lower() == 'none'):
+            self.set_defaults()
+            self.user = user
+            self.password = password
+            return
+
+        if not isinstance(s, str):
+            raise ValueError('proxy config not a string')
+
+        args = s.split(':')
+        if args[0] in ProxySettings.MODES:
+            self.mode = args[0]
+            args = args[1:]
+
+        # detect migrate from old settings
+        if len(args) == 4 and is_valid_host(args[0]) and is_valid_port(args[1]):  # host:port:user:pass,
+            self.host = args[0]
+            self.port = args[1]
+            self.user = args[2]
+            self.password = args[3]
+        else:
+            self.host = ':'.join(args[:-1])
+            self.port = args[-1]
+            self.user = user
+            self.password = password
+
+        if not is_valid_host(self.host) or not is_valid_port(self.port):
+            self.enabled = False
+
+    def to_dict(self):
+        return {
+            'enabled': self.enabled,
+            'mode': self.mode,
+            'host': self.host,
+            'port': self.port,
+            'user': self.user,
+            'password': self.password
+        }
+
+    @classmethod
+    def from_config(cls, config: 'SimpleConfig') -> 'ProxySettings':
+        proxy = ProxySettings()
+        proxy.deserialize_proxy_cfgstr(
+            config.NETWORK_PROXY, config.NETWORK_PROXY_USER, config.NETWORK_PROXY_PASSWORD
+        )
+        proxy.enabled = config.NETWORK_PROXY_ENABLED
+        return proxy
+
+    @classmethod
+    def from_dict(cls, d: dict) -> 'ProxySettings':
+        proxy = ProxySettings()
+        proxy.enabled = d.get('enabled', proxy.enabled)
+        proxy.mode = d.get('mode', proxy.mode)
+        proxy.host = d.get('host', proxy.host)
+        proxy.port = d.get('port', proxy.port)
+        proxy.user = d.get('user', proxy.user)
+        proxy.password = d.get('password', proxy.password)
+        return proxy
+
+    @classmethod
+    def probe_tor(cls, on_finished: Callable[[str | None, int | None], None]):
+        async def detect_task(finished: Callable[[str | None, int | None], None]):
+            try:
+                net_addr = await detect_tor_socks_proxy()
+                if net_addr is None:
+                    finished('', -1)
+                else:
+                    host = net_addr[0]
+                    port = net_addr[1]
+                    finished(host, port)
+            finally:
+                cls.probe_fut = None
+
+        if cls.probe_fut:  # one probe at a time
+            return
+        cls.probe_fut = asyncio.run_coroutine_threadsafe(detect_task(on_finished), util.get_asyncio_loop())
+
+    def __eq__(self, other):
+        return self.enabled == other.enabled \
+            and self.mode == other.mode \
+            and self.host == other.host \
+            and self.port == other.port \
+            and self.user == other.user \
+            and self.password == other.password
+
+    def __str__(self):
+        return f'{self.enabled=} {self.mode=} {self.host=} {self.port=} {self.user=}'
+
+
 class NetworkParameters(NamedTuple):
     server: ServerAddr
-    proxy: Optional[dict]
+    proxy: ProxySettings
     auto_connect: bool
     oneserver: bool = False
 
 
-proxy_modes = ['socks4', 'socks5']
-
-
-def serialize_proxy(p):
-    if not isinstance(p, dict):
-        return None
-    return ':'.join([p.get('mode'), p.get('host'), p.get('port'),
-                     p.get('user', ''), p.get('password', '')])
-
-
-def deserialize_proxy(s: str) -> Optional[dict]:
-    if not isinstance(s, str):
-        return None
-    if s.lower() == 'none':
-        return None
-    proxy = {"mode":"socks5", "host":"localhost"}
-    # FIXME raw IPv6 address fails here
-    args = s.split(':')
-    n = 0
-    if proxy_modes.count(args[n]) == 1:
-        proxy["mode"] = args[n]
-        n += 1
-    if len(args) > n:
-        proxy["host"] = args[n]
-        n += 1
-    if len(args) > n:
-        proxy["port"] = args[n]
-        n += 1
-    else:
-        proxy["port"] = "8080" if proxy["mode"] == "http" else "1080"
-    if len(args) > n:
-        proxy["user"] = args[n]
-        n += 1
-    if len(args) > n:
-        proxy["password"] = args[n]
-    return proxy
-
-
 class BestEffortRequestFailed(NetworkException): pass
 
 
@@ -238,14 +322,21 @@ class UntrustedServerReturnedError(NetworkException):
     def get_message_for_gui(self) -> str:
         return str(self)
 
-    def __str__(self):
-        return _("The server returned an error.")
-
-    def __repr__(self):
+    def get_untrusted_message(self) -> str:
         e = self.original_exception
         return (f"<UntrustedServerReturnedError "
                 f"[DO NOT TRUST THIS MESSAGE] original_exception: {error_text_str_to_safe_str(repr(e))}>")
 
+    def __str__(self):
+        # We should not show the untrusted text from self.original_exception,
+        # to avoid accidentally showing it in the GUI.
+        return _("The server returned an error.")
+
+    def __repr__(self):
+        # We should not show the untrusted text from self.original_exception,
+        # to avoid accidentally showing it in the GUI.
+        return f"<UntrustedServerReturnedError {str(self)!r}>"
+
 
 _INSTANCE = None
 
@@ -255,8 +346,6 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
     servers, each connected socket is handled by an Interface() object.
     """
 
-    LOGGING_SHORTCUT = 'n'
-
     taskgroup: Optional[OldTaskGroup]
     interface: Optional[Interface]
     interfaces: Dict[ServerAddr, Interface]
@@ -267,10 +356,9 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
 
     channel_db: Optional['ChannelDB'] = None
     lngossip: Optional['LNGossip'] = None
-    local_watchtower: Optional['WatchTower'] = None
     path_finder: Optional['LNPathFinder'] = None
 
-    def __init__(self, config: SimpleConfig, *, daemon: 'Daemon' = None):
+    def __init__(self, config: 'SimpleConfig', *, daemon: 'Daemon' = None):
         global _INSTANCE
         assert _INSTANCE is None, "Network is a singleton!"
         _INSTANCE = self
@@ -287,21 +375,21 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
         self.asyncio_loop = util.get_asyncio_loop()
         assert self.asyncio_loop.is_running(), "event loop not running"
 
-        assert isinstance(config, SimpleConfig), f"config should be a SimpleConfig instead of {type(config)}"
         self.config = config
-
         self.daemon = daemon
 
         blockchain.read_blockchains(self.config)
         blockchain.init_headers_file_for_best_chain()
         self.logger.info(f"blockchains {list(map(lambda b: b.forkpoint, blockchain.blockchains.values()))}")
-        self._blockchain_preferred_block = self.config.get('blockchain_preferred_block', None)  # type: Dict[str, Any]
+        self._blockchain_preferred_block = self.config.BLOCKCHAIN_PREFERRED_BLOCK  # type: Dict[str, Any]
         if self._blockchain_preferred_block is None:
             self._set_preferred_chain(None)
         self._blockchain = blockchain.get_best_chain()
 
         self._allowed_protocols = {PREFERRED_NETWORK_PROTOCOL}
 
+        self.proxy = ProxySettings()
+        self.is_proxy_tor = None  # type: Optional[bool]  # tri-state. None means unknown.
         self._init_parameters_from_config()
 
         self.taskgroup = None
@@ -341,11 +429,9 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
         self._has_ever_managed_to_connect_to_server = False
         self._was_started = False
 
-        # lightning network
-        if self.config.get('run_watchtower', False):
-            from . import lnwatcher
-            self.local_watchtower = lnwatcher.WatchTower(self)
-            asyncio.ensure_future(self.local_watchtower.start_watching())
+        self.mempool_fees = FeeHistogram()
+        self.fee_estimates = FeeTimeEstimates()
+        self.last_time_fee_estimates_requested = 0  # zero ensures immediate fees
 
     def has_internet_connection(self) -> bool:
         """Our guess whether the device has Internet-connectivity."""
@@ -358,13 +444,13 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
         from . import lnrouter
         from . import channel_db
         from . import lnworker
-        if not self.config.get('use_gossip'):
+        if not self.config.LIGHTNING_USE_GOSSIP:
             return
         if self.lngossip is None:
             self.channel_db = channel_db.ChannelDB(self)
             self.path_finder = lnrouter.LNPathFinder(self.channel_db)
             self.channel_db.load_data()
-            self.lngossip = lnworker.LNGossip()
+            self.lngossip = lnworker.LNGossip(self.config)
             self.lngossip.start_network(self)
 
     async def stop_gossip(self, *, full_shutdown: bool = False):
@@ -386,6 +472,9 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
 
     @staticmethod
     def get_instance() -> Optional["Network"]:
+        """Return the global singleton network instance.
+        Note that this can return None! If we are run with the --offline flag, there is no network.
+        """
         return _INSTANCE
 
     def with_recent_servers_lock(func):
@@ -455,8 +544,10 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
         async def get_banner():
             self.banner = await interface.get_server_banner()
             util.trigger_callback('banner', self.banner)
+
         async def get_donation_address():
             self.donation_address = await interface.get_donation_address()
+
         async def get_server_peers():
             server_peers = await session.send_request('server.peers.subscribe')
             random.shuffle(server_peers)
@@ -465,6 +556,7 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
             # note that 'parse_servers' also validates the data (which is untrusted input!)
             self.server_peers = parse_servers(server_peers)
             util.trigger_callback('servers', self.get_servers())
+
         async def get_relay_fee():
             self.relay_fee = await interface.get_relay_fee()
 
@@ -476,11 +568,27 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
             await group.spawn(self._request_fee_estimates(interface))
 
     async def _request_fee_estimates(self, interface):
-        self.config.requested_fee_estimates()
+        self.requested_fee_estimates()
         histogram = await interface.get_fee_histogram()
-        self.config.mempool_fees = histogram
-        self.logger.info(f'fee_histogram {histogram}')
-        util.trigger_callback('fee_histogram', self.config.mempool_fees)
+        self.mempool_fees.set_data(histogram)
+        self.logger.info(f'fee_histogram {len(histogram)}')
+        util.trigger_callback('fee_histogram', self.mempool_fees)
+
+    def is_fee_estimates_update_required(self):
+        """Checks time since last requested and updated fee estimates.
+        Returns True if an update should be requested.
+        """
+        now = time.time()
+        return now - self.last_time_fee_estimates_requested > 60
+
+    def has_fee_etas(self):
+        return self.fee_estimates.has_data()
+
+    def has_fee_mempool(self) -> bool:
+        return self.mempool_fees.has_data()
+
+    def requested_fee_estimates(self):
+        self.last_time_fee_estimates_requested = time.time()
 
     def get_parameters(self) -> NetworkParameters:
         return NetworkParameters(server=self.default_server,
@@ -489,9 +597,17 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
                                  oneserver=self.oneserver)
 
     def _init_parameters_from_config(self) -> None:
-        self.auto_connect = self.config.get('auto_connect', True)
+        dns_hacks.configure_dns_resolver()
+        self.auto_connect = self.config.NETWORK_AUTO_CONNECT
+        if self.auto_connect and self.config.NETWORK_ONESERVER:
+            # enabling both oneserver and auto_connect doesn't really make sense
+            # assume oneserver is enabled for privacy reasons, disable auto_connect and assume server is unpredictable
+            self.logger.warning(f'both "oneserver" and "auto_connect" options enabled, disabling "auto_connect" and resetting "server".')
+            self.config.NETWORK_SERVER = ""  # let _set_default_server set harmless default (localhost)
+            self.auto_connect = False
+
         self._set_default_server()
-        self._set_proxy(deserialize_proxy(self.config.get('proxy')))
+        self._set_proxy(ProxySettings.from_config(self.config))
         self._maybe_set_oneserver()
 
     def get_donation_address(self):
@@ -509,11 +625,10 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
 
     def get_fee_estimates(self):
         from statistics import median
-        from .simple_config import FEE_ETA_TARGETS
         if self.auto_connect:
             with self.interfaces_lock:
                 out = {}
-                for n in FEE_ETA_TARGETS:
+                for n in FEE_ETA_TARGETS[0:-1]:
                     try:
                         out[n] = int(median(filter(None, [i.fee_estimates_eta.get(n) for i in self.interfaces.values()])))
                     except Exception:
@@ -528,11 +643,12 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
         if fee_est is None:
             fee_est = self.get_fee_estimates()
         for nblock_target, fee in fee_est.items():
-            self.config.update_fee_estimates(nblock_target, fee)
+            self.fee_estimates.set_data(nblock_target, fee)
         if not hasattr(self, "_prev_fee_est") or self._prev_fee_est != fee_est:
             self._prev_fee_est = copy.copy(fee_est)
             self.logger.info(f'fee_estimates {fee_est}')
-        util.trigger_callback('fee', self.config.fee_estimates)
+        util.trigger_callback('fee', self.fee_estimates)
+
 
     @with_recent_servers_lock
     def get_servers(self):
@@ -553,8 +669,20 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
                 out[server.host].update({server.protocol: port})
             else:
                 out[server.host] = {server.protocol: port}
+        # add bookmarks
+        bookmarks = self.config.NETWORK_BOOKMARKED_SERVERS or []
+        for server_str in bookmarks:
+            try:
+                server = ServerAddr.from_str(server_str)
+            except ValueError:
+                continue
+            port = str(server.port)
+            if server.host in out:
+                out[server.host].update({server.protocol: port})
+            else:
+                out[server.host] = {server.protocol: port}
         # potentially filter out some
-        if self.config.get('noonion'):
+        if self.config.NETWORK_NOONION:
             out = filter_noonion(out)
         return out
 
@@ -590,7 +718,7 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
 
     def _set_default_server(self) -> None:
         # Server for addresses and transactions
-        server = self.config.get('server', None)
+        server = self.config.NETWORK_SERVER
         # Sanitize default server
         if server:
             try:
@@ -599,43 +727,73 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
                 self.logger.warning(f'failed to parse server-string ({server!r}); falling back to localhost:1:s.')
                 self.default_server = ServerAddr.from_str("localhost:1:s")
         else:
-            self.default_server = pick_random_server(allowed_protocols=self._allowed_protocols)
+            # if oneserver is enabled but no server specified then don't pick a random server
+            if self.config.NETWORK_ONESERVER:
+                self.logger.warning(f'"oneserver" option enabled, but no "server" defined; falling back to localhost:1:s.')
+                self.default_server = ServerAddr.from_str("localhost:1:s")
+            else:
+                self.default_server = pick_random_server(allowed_protocols=self._allowed_protocols)
         assert isinstance(self.default_server, ServerAddr), f"invalid type for default_server: {self.default_server!r}"
 
-    def _set_proxy(self, proxy: Optional[dict]):
-        self.proxy = proxy
-        dns_hacks.configure_dns_depending_on_proxy(bool(proxy))
+    def _set_proxy(self, proxy: ProxySettings):
+        if self.proxy == proxy:
+            return
+
         self.logger.info(f'setting proxy {proxy}')
+        self.proxy = proxy
 
-        self.tor_proxy = False
-        if bool(proxy) and proxy['mode'] == 'socks5':
-            # test for Tor
-            self.tor_proxy = util.is_tor_socks_port(proxy['host'], int(proxy['port']))
-            if self.tor_proxy:
-                self.logger.info(f'Proxy is TOR')
+        # reset is_proxy_tor to unknown, and re-detect it:
+        self.is_proxy_tor = None
+        self._detect_if_proxy_is_tor()
 
-        util.trigger_callback('proxy_set', self.proxy, self.tor_proxy)
+        util.trigger_callback('proxy_set', self.proxy)
+
+    def _detect_if_proxy_is_tor(self) -> None:
+        async def tor_probe_task(p):
+            assert p is not None
+            is_tor = await util.is_tor_socks_port(p.host, int(p.port))
+            if self.proxy == p:  # is this the proxy we probed?
+                if self.is_proxy_tor != is_tor:
+                    self.logger.info(f'Proxy is {"" if is_tor else "not "}TOR')
+                    self.is_proxy_tor = is_tor
+                util.trigger_callback('tor_probed', is_tor)
+
+        proxy = self.proxy
+        if proxy and proxy.enabled and proxy.mode == 'socks5':
+            asyncio.run_coroutine_threadsafe(tor_probe_task(proxy), self.asyncio_loop)
 
     @log_exceptions
     async def set_parameters(self, net_params: NetworkParameters):
         proxy = net_params.proxy
-        proxy_str = serialize_proxy(proxy)
+        proxy_str = proxy.serialize_proxy_cfgstr()
+        proxy_enabled = proxy.enabled
+        proxy_user = proxy.user
+        proxy_pass = proxy.password
         server = net_params.server
         # sanitize parameters
         try:
             if proxy:
-                proxy_modes.index(proxy['mode']) + 1
-                int(proxy['port'])
+                # proxy_modes.index(proxy['mode']) + 1
+                ProxySettings.MODES.index(proxy.mode) + 1
+                # int(proxy['port'])
+                int(proxy.port)
         except Exception:
-            return
-        self.config.set_key('auto_connect', net_params.auto_connect, False)
-        self.config.set_key('oneserver', net_params.oneserver, False)
-        self.config.set_key('proxy', proxy_str, False)
-        self.config.set_key('server', str(server), True)
+            proxy.enabled = False
+            # return
+        self.config.NETWORK_AUTO_CONNECT = net_params.auto_connect
+        self.config.NETWORK_ONESERVER = net_params.oneserver
+        self.config.NETWORK_PROXY_ENABLED = proxy_enabled
+        self.config.NETWORK_PROXY = proxy_str
+        self.config.NETWORK_PROXY_USER = proxy_user
+        self.config.NETWORK_PROXY_PASSWORD = proxy_pass
+        self.config.NETWORK_SERVER = str(server)
         # abort if changes were not allowed by config
-        if self.config.get('server') != str(server) \
-                or self.config.get('proxy') != proxy_str \
-                or self.config.get('oneserver') != net_params.oneserver:
+        if self.config.NETWORK_SERVER != str(server) \
+                or self.config.NETWORK_PROXY_ENABLED != proxy_enabled \
+                or self.config.NETWORK_PROXY != proxy_str \
+                or self.config.NETWORK_PROXY_USER != proxy_user \
+                or self.config.NETWORK_PROXY_PASSWORD != proxy_pass \
+                or self.config.NETWORK_ONESERVER != net_params.oneserver:
             return
 
         proxy_changed = self.proxy != proxy
@@ -657,10 +815,26 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
         util.trigger_callback('network_updated')
 
     def _maybe_set_oneserver(self) -> None:
-        oneserver = bool(self.config.get('oneserver', False))
+        oneserver = self.config.NETWORK_ONESERVER
         self.oneserver = oneserver
         self.num_server = NUM_TARGET_CONNECTED_SERVERS if not oneserver else 0
 
+    def is_server_bookmarked(self, server: ServerAddr) -> bool:
+        bookmarks = self.config.NETWORK_BOOKMARKED_SERVERS or []
+        return str(server) in bookmarks
+
+    def set_server_bookmark(self, server: ServerAddr, *, add: bool) -> None:
+        server_str = str(server)
+        with self.config.lock:
+            bookmarks = self.config.NETWORK_BOOKMARKED_SERVERS or []
+            if add:
+                if server_str not in bookmarks:
+                    bookmarks.append(server_str)
+            else:  # remove
+                if server_str in bookmarks:
+                    bookmarks.remove(server_str)
+            self.config.NETWORK_BOOKMARKED_SERVERS = bookmarks
+
     async def _switch_to_random_interface(self):
         '''Switch to a random connected server other than the current one'''
         servers = self.get_interfaces()    # Those in connected state
@@ -788,11 +962,11 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
         util.trigger_callback('network_updated')
 
     def get_network_timeout_seconds(self, request_type=NetworkTimeout.Generic) -> int:
-        if self.config.get('network_timeout', None):
-            return int(self.config.get('network_timeout'))
+        if self.config.NETWORK_TIMEOUT:
+            return self.config.NETWORK_TIMEOUT
         if self.oneserver and not self.auto_connect:
             return request_type.MOST_RELAXED
-        if self.proxy:
+        if self.proxy and self.proxy.enabled:
             return request_type.RELAXED
         return request_type.NORMAL
 
@@ -809,11 +983,11 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
             self._set_status(ConnectionState.CONNECTING)
         self._trying_addr_now(server)
 
-        interface = Interface(network=self, server=server, proxy=self.proxy)
+        interface = Interface(network=self, server=server)
         # note: using longer timeouts here as DNS can sometimes be slow!
         timeout = self.get_network_timeout_seconds(NetworkTimeout.Generic)
         try:
-            await asyncio.wait_for(interface.ready, timeout)
+            await util.wait_for2(interface.ready, timeout)
         except BaseException as e:
             self.logger.info(f"couldn't launch iface {server} -- {repr(e)}")
             await interface.close()
@@ -831,6 +1005,9 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
         self._has_ever_managed_to_connect_to_server = True
         self._add_recent_server(server)
         util.trigger_callback('network_updated')
+        # When the proxy settings were set, the proxy (if any) might have been unreachable,
+        # resulting in a false-negative for Tor-detection. Given we just connected to a server, re-test now.
+        self._detect_if_proxy_is_tor()
 
     def check_interface_against_healthy_spread_of_connected_servers(self, iface_to_check: Interface) -> bool:
         # main interface is exempt. this makes switching servers easier
@@ -891,12 +1068,18 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
         return make_reliable_wrapper
 
     def catch_server_exceptions(func):
+        """Decorator that wraps server errors in UntrustedServerReturnedError,
+        to avoid showing untrusted arbitrary text to users.
+        """
         @functools.wraps(func)
         async def wrapper(self, *args, **kwargs):
             try:
                 return await func(self, *args, **kwargs)
             except aiorpcx.jsonrpc.CodeMessageError as e:
-                raise UntrustedServerReturnedError(original_exception=e) from e
+                wrapped_exc = UntrustedServerReturnedError(original_exception=e)
+                # log (sanitized) untrusted error text now, to ease debugging
+                self.logger.debug(f"got error from server for {func.__qualname__}: {wrapped_exc.get_untrusted_message()!r}")
+                raise wrapped_exc from e
         return wrapper
 
     @best_effort_reliable
@@ -908,25 +1091,28 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
 
     @best_effort_reliable
     async def broadcast_transaction(self, tx: 'Transaction', *, timeout=None) -> None:
+        """caller should handle TxBroadcastError"""
         if self.interface is None:  # handled by best_effort_reliable
             raise RequestTimedOut()
         if timeout is None:
             timeout = self.get_network_timeout_seconds(NetworkTimeout.Urgent)
+        if any(DummyAddress.is_dummy_address(txout.address) for txout in tx.outputs()):
+            raise DummyAddressUsedInTxException("tried to broadcast tx with dummy address!")
         try:
             out = await self.interface.session.send_request('blockchain.transaction.broadcast', [tx.serialize()], timeout=timeout)
             # note: both 'out' and exception messages are untrusted input from the server
         except (RequestTimedOut, asyncio.CancelledError, asyncio.TimeoutError):
             raise  # pass-through
         except aiorpcx.jsonrpc.CodeMessageError as e:
-            self.logger.info(f"broadcast_transaction error [DO NOT TRUST THIS MESSAGE]: {error_text_str_to_safe_str(repr(e))}")
+            self.logger.info(f"broadcast_transaction error [DO NOT TRUST THIS MESSAGE]: {error_text_str_to_safe_str(repr(e))}. tx={str(tx)}")
             raise TxBroadcastServerReturnedError(self.sanitize_tx_broadcast_response(e.message)) from e
         except BaseException as e:  # intentional BaseException for sanity!
-            self.logger.info(f"broadcast_transaction error2 [DO NOT TRUST THIS MESSAGE]: {error_text_str_to_safe_str(repr(e))}")
+            self.logger.info(f"broadcast_transaction error2 [DO NOT TRUST THIS MESSAGE]: {error_text_str_to_safe_str(repr(e))}. tx={str(tx)}")
             send_exception_to_crash_reporter(e)
             raise TxBroadcastUnknownError() from e
         if out != tx.txid():
             self.logger.info(f"unexpected txid for broadcast_transaction [DO NOT TRUST THIS MESSAGE]: "
-                             f"{error_text_str_to_safe_str(out)} != {tx.txid()}")
+                             f"{error_text_str_to_safe_str(out)} != {tx.txid()}. tx={str(tx)}")
             raise TxBroadcastHashMismatch(_("Server returned unexpected transaction ID."))
 
     async def try_broadcasting(self, tx, name) -> bool:
@@ -1046,18 +1232,25 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
                 return msg if msg else substring
         # https://github.com/bitcoin/bitcoin/blob/5bb64acd9d3ced6e6f95df282a1a0f8b98522cb0/src/rpc/rawtransaction.cpp
         # https://github.com/bitcoin/bitcoin/blob/5bb64acd9d3ced6e6f95df282a1a0f8b98522cb0/src/util/error.cpp
+        # https://github.com/bitcoin/bitcoin/blob/3f83c744ac28b700090e15b5dda2260724a56f49/src/common/messages.cpp#L126
         # grep "RPC_TRANSACTION"
         # grep "RPC_DESERIALIZATION_ERROR"
+        # grep "TransactionError"
         rawtransaction_error_messages = {
             r"Missing inputs": None,
             r"Inputs missing or spent": None,
             r"transaction already in block chain": None,
             r"Transaction already in block chain": None,
+            r"Transaction outputs already in utxo set": None,
             r"TX decode failed": None,
             r"Peer-to-peer functionality missing or disabled": None,
             r"Transaction rejected by AcceptToMemoryPool": None,
             r"AcceptToMemoryPool failed": None,
+            r"Transaction rejected by mempool": None,
+            r"Mempool internal error": None,
             r"Fee exceeds maximum configured by user": None,
+            r"Unspendable output exceeds maximum configured by user": None,
+            r"Transaction rejected due to invalid package": None,
         }
         for substring in rawtransaction_error_messages:
             if substring in server_msg:
@@ -1118,7 +1311,13 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
 
     @best_effort_reliable
     @catch_server_exceptions
-    async def request_chunk(self, height: int, tip=None, *, can_return_early=False):
+    async def request_chunk(
+        self,
+        height: int,
+        *,
+        tip: Optional[int] = None,
+        can_return_early: bool = False,
+    ) -> Optional[Tuple[bool, int]]:
         if self.interface is None:  # handled by best_effort_reliable
             raise RequestTimedOut()
         return await self.interface.request_chunk(height, tip=tip, can_return_early=can_return_early)
@@ -1185,7 +1384,7 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
             'height': height,
             'hash': header_hash,
         }
-        self.config.set_key('blockchain_preferred_block', self._blockchain_preferred_block)
+        self.config.BLOCKCHAIN_PREFERRED_BLOCK = self._blockchain_preferred_block
 
     async def follow_chain_given_id(self, chain_id: str) -> None:
         bc = blockchain.blockchains.get(chain_id)
@@ -1218,7 +1417,7 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
         interface = self.interface
         return interface.tip if interface else 0
 
-    def get_local_height(self):
+    def get_local_height(self) -> int:
         """Length of header chain, POW-verified.
         In case of a chain split, this is for the branch the main interface is on,
         but it is the tip of that branch (even if main interface is behind).
@@ -1321,7 +1520,7 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
         async def maintain_main_interface():
             await self._ensure_there_is_a_main_interface()
             if self.is_connected():
-                if self.config.is_fee_estimates_update_required():
+                if self.is_fee_estimates_update_required():
                     await self.interface.taskgroup.spawn(self._request_fee_estimates, self.interface)
 
         while True:
@@ -1366,12 +1565,8 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
 
     @classmethod
     def send_http_on_proxy(cls, method, url, **kwargs):
-        network = cls.get_instance()
-        if network:
-            assert util.get_running_loop() != network.asyncio_loop
-            loop = network.asyncio_loop
-        else:
-            loop = util.get_asyncio_loop()
+        loop = util.get_asyncio_loop()
+        assert util.get_running_loop() != loop, 'must not be called from asyncio thread'
         coro = asyncio.run_coroutine_threadsafe(cls.async_send_http_on_proxy(method, url, **kwargs), loop)
         # note: _send_http_on_proxy has its own timeout, so no timeout here:
         return coro.result()
@@ -1395,9 +1590,9 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
             timeout = self.get_network_timeout_seconds(NetworkTimeout.Urgent)
         responses = dict()
         async def get_response(server: ServerAddr):
-            interface = Interface(network=self, server=server, proxy=self.proxy)
+            interface = Interface(network=self, server=server)
             try:
-                await asyncio.wait_for(interface.ready, timeout)
+                await util.wait_for2(interface.ready, timeout)
             except BaseException as e:
                 await interface.close()
                 return
