@@ -9,10 +9,11 @@ from collections import defaultdict
 import logging
 import concurrent
 from concurrent import futures
-import unittest
+from unittest import mock
 from typing import Iterable, NamedTuple, Tuple, List, Dict
 
 from aiorpcx import timeout_after, TaskTimeout
+from electrum_ecc import ECPrivkey
 
 import electrum
 import electrum.trampoline
@@ -20,14 +21,15 @@ from electrum import bitcoin
 from electrum import util
 from electrum import constants
 from electrum.network import Network
-from electrum.ecc import ECPrivkey
 from electrum import simple_config, lnutil
 from electrum.lnaddr import lnencode, LnAddr, lndecode
 from electrum.bitcoin import COIN, sha256
+from electrum.transaction import Transaction
 from electrum.util import NetworkRetryManager, bfh, OldTaskGroup, EventListener, InvoiceError
 from electrum.lnpeer import Peer
-from electrum.lnutil import LNPeerAddr, Keypair, privkey_to_pubkey
-from electrum.lnutil import PaymentFailure, LnFeatures, HTLCOwner, PaymentFeeBudget
+from electrum.lntransport import LNPeerAddr
+from electrum.crypto import privkey_to_pubkey
+from electrum.lnutil import Keypair, PaymentFailure, LnFeatures, HTLCOwner, PaymentFeeBudget
 from electrum.lnchannel import ChannelState, PeerState, Channel
 from electrum.lnrouter import LNPathFinder, PathEdge, LNPathInconsistent
 from electrum.channel_db import ChannelDB
@@ -39,13 +41,17 @@ from electrum.lnworker import PaymentInfo, RECEIVED
 from electrum.lnonion import OnionFailureCode, OnionRoutingFailure
 from electrum.lnutil import UpdateAddHtlc
 from electrum.lnutil import LOCAL, REMOTE
-from electrum.invoices import PR_PAID, PR_UNPAID
+from electrum.invoices import PR_PAID, PR_UNPAID, Invoice
 from electrum.interface import GracefulDisconnect
 from electrum.simple_config import SimpleConfig
+from electrum.fee_policy import FeeTimeEstimates, FEE_ETA_TARGETS
+from electrum.mpp_split import split_amount_normal
 
 from .test_lnchannel import create_test_channels
+from .test_bitcoin import needs_test_with_all_chacha20_implementations
 from . import ElectrumTestCase
 
+
 def keypair():
     priv = ECPrivkey.generate_random_key().get_secret_bytes()
     k1 = Keypair(
@@ -62,6 +68,8 @@ class MockNetwork:
         self.callbacks = defaultdict(list)
         self.lnwatcher = None
         self.interface = None
+        self.fee_estimates = FeeTimeEstimates()
+        self.populate_fee_estimates()
         self.config = config
         self.asyncio_loop = util.get_asyncio_loop()
         self.channel_db = ChannelDB(self)
@@ -88,6 +96,10 @@ class MockNetwork:
     async def try_broadcasting(self, tx, name):
         await self.broadcast_transaction(tx)
 
+    def populate_fee_estimates(self):
+        for target in FEE_ETA_TARGETS[:-1]:
+            self.fee_estimates.set_data(target, 50000 // target)
+
 
 class MockBlockchain:
 
@@ -101,13 +113,20 @@ class MockBlockchain:
 
 
 class MockADB:
+    def __init__(self):
+        self._blockchain = MockBlockchain()
     def add_transaction(self, tx):
         pass
+    def get_local_height(self):
+        return self._blockchain.height()
 
 class MockWallet:
     receive_requests = {}
     adb = MockADB()
 
+    def get_invoice(self, key):
+        pass
+
     def get_request(self, key):
         pass
 
@@ -129,6 +148,13 @@ class MockWallet:
     def get_fingerprint(self):
         return ''
 
+    def get_new_sweep_address_for_channel(self):
+        # note: sweep is not tested here, only in regtest
+        return "tb1qqu5newtapamjchgxf0nty6geuykhvwas45q4q4"
+
+    def is_up_to_date(self):
+        return True
+
 
 class MockLNGossip:
     def get_sync_progress_estimate(self):
@@ -142,18 +168,19 @@ class MockLNWallet(Logger, EventListener, NetworkRetryManager[LNPeerAddr]):
     MPP_SPLIT_PART_FRACTION = 1  # this disables the forced splitting
     MPP_SPLIT_PART_MINAMT_MSAT = 5_000_000
 
-    def __init__(self, *, local_keypair: Keypair, chans: Iterable['Channel'], tx_queue, name):
+    def __init__(self, *, local_keypair: Keypair, chans: Iterable['Channel'], tx_queue, name, has_anchors):
         self.name = name
         Logger.__init__(self)
         NetworkRetryManager.__init__(self, max_retry_delay_normal=1, init_retry_delay_normal=1)
         self.node_keypair = local_keypair
-        self.payment_secret_key = os.urandom(256) # does not need to be deterministic in tests
+        self.payment_secret_key = os.urandom(32)  # does not need to be deterministic in tests
         self._user_dir = tempfile.mkdtemp(prefix="electrum-lnpeer-test-")
         self.config = SimpleConfig({}, read_user_dir_function=lambda: self._user_dir)
         self.network = MockNetwork(tx_queue, config=self.config)
         self.taskgroup = OldTaskGroup()
         self.lnwatcher = None
         self.swap_manager = None
+        self.onion_message_manager = None
         self.listen_server = None
         self._channels = {chan.channel_id: chan for chan in chans}
         self.payment_info = {}
@@ -167,6 +194,8 @@ class MockLNWallet(Logger, EventListener, NetworkRetryManager[LNPeerAddr]):
         self.features |= LnFeatures.OPTION_TRAMPOLINE_ROUTING_OPT_ELECTRUM
         self.features |= LnFeatures.OPTION_CHANNEL_TYPE_OPT
         self.features |= LnFeatures.OPTION_SCID_ALIAS_OPT
+        self.features |= LnFeatures.OPTION_STATIC_REMOTEKEY_OPT
+        self.config.ENABLE_ANCHOR_CHANNELS = has_anchors
         self.pending_payments = defaultdict(asyncio.Future)
         for chan in chans:
             chan.lnworker = self
@@ -181,9 +210,10 @@ class MockLNWallet(Logger, EventListener, NetworkRetryManager[LNPeerAddr]):
         self.active_forwardings = {}
         self.forwarding_failures = {}
         self.inflight_payments = set()
-        self.preimages = {}
+        self._preimages = {}
         self.stopping_soon = False
         self.downstream_to_upstream_htlc = {}
+        self.dont_settle_htlcs = {}
         self.hold_invoice_callbacks = {}
         self.payment_bundles = [] # lists of hashes. todo:persist
         self.config.INITIAL_TRAMPOLINE_FEE_LEVEL = 0
@@ -276,7 +306,7 @@ class MockLNWallet(Logger, EventListener, NetworkRetryManager[LNPeerAddr]):
     get_preimage = LNWallet.get_preimage
     create_route_for_single_htlc = LNWallet.create_route_for_single_htlc
     create_routes_for_payment = LNWallet.create_routes_for_payment
-    _check_invoice = LNWallet._check_invoice
+    _check_bolt11_invoice = LNWallet._check_bolt11_invoice
     pay_to_route = LNWallet.pay_to_route
     pay_to_node = LNWallet.pay_to_node
     pay_invoice = LNWallet.pay_invoice
@@ -297,7 +327,7 @@ class MockLNWallet(Logger, EventListener, NetworkRetryManager[LNPeerAddr]):
     is_forwarded_htlc = LNWallet.is_forwarded_htlc
     notify_upstream_peer = LNWallet.notify_upstream_peer
     _force_close_channel = LNWallet._force_close_channel
-    suggest_splits = LNWallet.suggest_splits
+    suggest_payment_splits = LNWallet.suggest_payment_splits
     register_hold_invoice = LNWallet.register_hold_invoice
     unregister_hold_invoice = LNWallet.unregister_hold_invoice
     add_payment_info_for_hold_invoice = LNWallet.add_payment_info_for_hold_invoice
@@ -305,6 +335,7 @@ class MockLNWallet(Logger, EventListener, NetworkRetryManager[LNPeerAddr]):
     update_mpp_with_received_htlc = LNWallet.update_mpp_with_received_htlc
     set_mpp_resolution = LNWallet.set_mpp_resolution
     is_mpp_amount_reached = LNWallet.is_mpp_amount_reached
+    get_mpp_amounts = LNWallet.get_mpp_amounts
     get_first_timestamp_of_mpp = LNWallet.get_first_timestamp_of_mpp
     bundle_payments = LNWallet.bundle_payments
     get_payment_bundle = LNWallet.get_payment_bundle
@@ -312,6 +343,9 @@ class MockLNWallet(Logger, EventListener, NetworkRetryManager[LNPeerAddr]):
     save_forwarding_failure = LNWallet.save_forwarding_failure
     get_forwarding_failure = LNWallet.get_forwarding_failure
     maybe_cleanup_forwarding = LNWallet.maybe_cleanup_forwarding
+    current_target_feerate_per_kw = LNWallet.current_target_feerate_per_kw
+    current_low_feerate_per_kw_srk_channel = LNWallet.current_low_feerate_per_kw_srk_channel
+    maybe_cleanup_mpp = LNWallet.maybe_cleanup_mpp
 
 
 class MockTransport:
@@ -382,7 +416,7 @@ low_fee_channel = {
 }
 
 depleted_channel = {
-    'local_balance_msat': 0,
+    'local_balance_msat': 330 * 1000, # local pays anchors
     'remote_balance_msat': 10 * bitcoin.COIN * 1000,
     'local_base_fee_msat': 1_000,
     'local_fee_rate_millionths': 1,
@@ -420,7 +454,42 @@ _GRAPH_DEFINITIONS = {
         },
         'dave': {
         },
-    }
+    },
+    'line_graph': {
+        'alice': {
+            'channels': {
+                'bob': low_fee_channel.copy(),
+            },
+        },
+        'bob': {  # Trampoline Forwarder
+            'channels': {
+                'carol': low_fee_channel.copy(),
+            },
+            'config': {
+                SimpleConfig.EXPERIMENTAL_LN_FORWARD_PAYMENTS: True,
+                SimpleConfig.EXPERIMENTAL_LN_FORWARD_TRAMPOLINE_PAYMENTS: True,
+            },
+        },
+        'carol': {
+            'channels': {
+                'dave': low_fee_channel.copy(),
+            },
+            'config': {
+                SimpleConfig.EXPERIMENTAL_LN_FORWARD_PAYMENTS: True,
+            },
+        },
+        'dave': {  # Trampoline Forwarder
+            'channels': {
+                'edward': low_fee_channel.copy(),
+            },
+            'config': {
+                SimpleConfig.EXPERIMENTAL_LN_FORWARD_PAYMENTS: True,
+                SimpleConfig.EXPERIMENTAL_LN_FORWARD_TRAMPOLINE_PAYMENTS: True,
+            },
+        },
+        'edward': {
+        },
+    },
 }
 
 
@@ -435,6 +504,20 @@ class PaymentTimeout(Exception): pass
 class SuccessfulTest(Exception): pass
 
 
+def inject_chan_into_gossipdb(*, channel_db: ChannelDB, graph: Graph, node1name: str, node2name: str) -> None:
+    chan_ann_raw = graph.channels[(node1name, node2name)].construct_channel_announcement_without_sigs()[0]
+    chan_ann_dict = decode_msg(chan_ann_raw)[1]
+    channel_db.add_channel_announcements(chan_ann_dict, trusted=True)
+
+    chan_upd1_raw = graph.channels[(node1name, node2name)].get_outgoing_gossip_channel_update()
+    chan_upd1_dict = decode_msg(chan_upd1_raw)[1]
+    channel_db.add_channel_update(chan_upd1_dict, verify=False)
+
+    chan_upd2_raw = graph.channels[(node2name, node1name)].get_outgoing_gossip_channel_update()
+    chan_upd2_dict = decode_msg(chan_upd2_raw)[1]
+    channel_db.add_channel_update(chan_upd2_dict, verify=False)
+
+
 class TestPeer(ElectrumTestCase):
     TESTNET = True
 
@@ -469,7 +552,7 @@ class TestPeer(ElectrumTestCase):
             payment_hash: bytes = None,
             invoice_features: LnFeatures = None,
             min_final_cltv_delta: int = None,
-    ) -> Tuple[LnAddr, str]:
+    ) -> Tuple[LnAddr, Invoice]:
         amount_btc = amount_msat/Decimal(COIN*1000)
         if payment_preimage is None and not payment_hash:
             payment_preimage = os.urandom(32)
@@ -504,7 +587,7 @@ class TestPeer(ElectrumTestCase):
         )
         invoice = lnencode(lnaddr1, w2.node_keypair.privkey)
         lnaddr2 = lndecode(invoice)  # unlike lnaddr1, this now has a pubkey set
-        return lnaddr2, invoice
+        return lnaddr2, Invoice.from_bech32(invoice)
 
     async def _activate_trampoline(self, w: MockLNWallet):
         if w.network.channel_db:
@@ -515,7 +598,7 @@ class TestPeer(ElectrumTestCase):
     def prepare_recipient(self, w2, payment_hash, test_hold_invoice, test_failure):
         if not test_hold_invoice and not test_failure:
             return
-        preimage = bytes.fromhex(w2.preimages.pop(payment_hash.hex()))
+        preimage = bytes.fromhex(w2._preimages.pop(payment_hash.hex()))
         if test_hold_invoice:
             async def cb(payment_hash):
                 if not test_failure:
@@ -541,8 +624,8 @@ class TestPeerDirect(TestPeer):
         bob_channel.storage['node_id'] = bob_channel.node_id
         t1, t2 = transport_pair(k1, k2, alice_channel.name, bob_channel.name)
         q1, q2 = asyncio.Queue(), asyncio.Queue()
-        w1 = MockLNWallet(local_keypair=k1, chans=[alice_channel], tx_queue=q1, name=bob_channel.name)
-        w2 = MockLNWallet(local_keypair=k2, chans=[bob_channel], tx_queue=q2, name=alice_channel.name)
+        w1 = MockLNWallet(local_keypair=k1, chans=[alice_channel], tx_queue=q1, name=bob_channel.name, has_anchors=self.TEST_ANCHOR_CHANNELS)
+        w2 = MockLNWallet(local_keypair=k2, chans=[bob_channel], tx_queue=q2, name=alice_channel.name, has_anchors=self.TEST_ANCHOR_CHANNELS)
         self._lnworkers_created.extend([w1, w2])
         p1 = PeerInTests(w1, k2.pubkey, t1)
         p2 = PeerInTests(w2, k1.pubkey, t2)
@@ -1182,7 +1265,7 @@ class TestPeerDirect(TestPeer):
         async def action():
             await util.wait_for2(p1.initialized, 1)
             await util.wait_for2(p2.initialized, 1)
-            await p1.send_warning(alice_channel.channel_id, 'be warned!', close_connection=True)
+            p1.send_warning(alice_channel.channel_id, 'be warned!', close_connection=True)
         gath = asyncio.gather(action(), p1._message_loop(), p2._message_loop(), p1.htlc_switch(), p2.htlc_switch())
         with self.assertRaises(GracefulDisconnect):
             await gath
@@ -1194,7 +1277,7 @@ class TestPeerDirect(TestPeer):
         async def action():
             await util.wait_for2(p1.initialized, 1)
             await util.wait_for2(p2.initialized, 1)
-            await p1.send_error(alice_channel.channel_id, 'some error happened!', force_close_channel=True)
+            p1.send_error(alice_channel.channel_id, 'some error happened!', force_close_channel=True)
             assert alice_channel.is_closed()
             gath.cancel()
         gath = asyncio.gather(action(), p1._message_loop(), p2._message_loop(), p1.htlc_switch(), p2.htlc_switch())
@@ -1206,9 +1289,9 @@ class TestPeerDirect(TestPeer):
 
         # create upfront shutdown script for bob, alice doesn't use upfront
         # shutdown script
-        bob_uss_pub = lnutil.privkey_to_pubkey(os.urandom(32))
+        bob_uss_pub = privkey_to_pubkey(os.urandom(32))
         bob_uss_addr = bitcoin.pubkey_to_address('p2wpkh', bob_uss_pub.hex())
-        bob_uss = bfh(bitcoin.address_to_script(bob_uss_addr))
+        bob_uss = bitcoin.address_to_script(bob_uss_addr)
 
         # bob commits to close to bob_uss
         alice_channel.config[HTLCOwner.REMOTE].upfront_shutdown_script = bob_uss
@@ -1217,10 +1300,8 @@ class TestPeerDirect(TestPeer):
         bob_channel.config[HTLCOwner.LOCAL].upfront_shutdown_script = b''
 
         p1, p2, w1, w2, q1, q2 = self.prepare_peers(alice_channel, bob_channel)
-        w1.network.config.FEE_EST_DYNAMIC = False
-        w2.network.config.FEE_EST_DYNAMIC = False
-        w1.network.config.FEE_EST_STATIC_FEERATE = 5000
-        w2.network.config.FEE_EST_STATIC_FEERATE = 1000
+        w1.network.config.FEE_POLICY = 'feerate:5000'
+        w2.network.config.FEE_POLICY = 'feerate:1000'
 
         async def test():
             async def close():
@@ -1228,7 +1309,7 @@ class TestPeerDirect(TestPeer):
                 await util.wait_for2(p2.initialized, 1)
                 # bob closes channel with different shutdown script
                 await p1.close_channel(alice_channel.channel_id)
-                gath.cancel()
+                self.fail("p1.close_channel should have raised above!")
 
             async def main_loop(peer):
                     async with peer.taskgroup as group:
@@ -1241,16 +1322,18 @@ class TestPeerDirect(TestPeer):
 
         with self.assertRaises(GracefulDisconnect):
             await test()
+        # check that neither party broadcast a closing tx (as it was not even signed)
+        self.assertEqual(0, q1.qsize())
+        self.assertEqual(0, q2.qsize())
 
+        # -- new scenario:
         # bob sends the same upfront_shutdown_script has he announced
         alice_channel.config[HTLCOwner.REMOTE].upfront_shutdown_script = bob_uss
         bob_channel.config[HTLCOwner.LOCAL].upfront_shutdown_script = bob_uss
 
         p1, p2, w1, w2, q1, q2 = self.prepare_peers(alice_channel, bob_channel)
-        w1.network.config.FEE_EST_DYNAMIC = False
-        w2.network.config.FEE_EST_DYNAMIC = False
-        w1.network.config.FEE_EST_STATIC_FEERATE = 5000
-        w2.network.config.FEE_EST_STATIC_FEERATE = 1000
+        w1.network.config.FEE_POLICY = 'feerate:5000'
+        w2.network.config.FEE_POLICY = 'feerate:1000'
 
         async def test():
             async def close():
@@ -1271,12 +1354,18 @@ class TestPeerDirect(TestPeer):
         with self.assertRaises(asyncio.CancelledError):
             await test()
 
+        # check if p1 has broadcast the closing tx, and if it pays to Bob's uss
+        self.assertEqual(1, q1.qsize())
+        closing_tx = q1.get_nowait()  # type: Transaction
+        self.assertEqual(2, len(closing_tx.outputs()))
+        self.assertEqual(1, len(closing_tx.get_output_idxs_from_address(bob_uss_addr)))
+
     async def test_channel_usage_after_closing(self):
         alice_channel, bob_channel = create_test_channels()
         p1, p2, w1, w2, q1, q2 = self.prepare_peers(alice_channel, bob_channel)
         lnaddr, pay_req = self.prepare_invoice(w2)
 
-        lnaddr = w1._check_invoice(pay_req)
+        lnaddr = w1._check_bolt11_invoice(pay_req.lightning_invoice)
         shi = (await w1.create_routes_from_invoice(lnaddr.get_amount_msat(), decoded_invoice=lnaddr))[0][0]
         route, amount_msat = shi.route, shi.amount_msat
         assert amount_msat == lnaddr.get_amount_msat()
@@ -1411,7 +1500,6 @@ class TestPeerForwarding(TestPeer):
         transports = {}
         workers = {}  # type: Dict[str, MockLNWallet]
         peers = {}
-
         # create channels
         for a, definition in graph_definition.items():
             for b, channel_def in definition.get('channels', {}).items():
@@ -1422,6 +1510,7 @@ class TestPeerForwarding(TestPeer):
                     bob_pubkey=keys[b].pubkey,
                     local_msat=channel_def['local_balance_msat'],
                     remote_msat=channel_def['remote_balance_msat'],
+                    anchor_outputs=self.TEST_ANCHOR_CHANNELS
                 )
                 channels[(a, b)], channels[(b, a)] = channel_ab, channel_ba
                 transport_ab, transport_ba = transport_pair(keys[a], keys[b], channel_ab.name, channel_ba.name)
@@ -1435,7 +1524,7 @@ class TestPeerForwarding(TestPeer):
         # create workers and peers
         for a, definition in graph_definition.items():
             channels_of_node = [c for k, c in channels.items() if k[0] == a]
-            workers[a] = MockLNWallet(local_keypair=keys[a], chans=channels_of_node, tx_queue=txs_queues[a], name=a)
+            workers[a] = MockLNWallet(local_keypair=keys[a], chans=channels_of_node, tx_queue=txs_queues[a], name=a, has_anchors=self.TEST_ANCHOR_CHANNELS)
         self._lnworkers_created.extend(list(workers.values()))
 
         # create peers
@@ -1472,6 +1561,55 @@ class TestPeerForwarding(TestPeer):
             print(f"       {keys[a].pubkey.hex()}")
         return graph
 
+    async def test_payment_in_graph_with_direct_channel(self):
+        """Test payment over a direct channel where sender has multiple available channels."""
+        graph = self.prepare_chans_and_peers_in_graph(self.GRAPH_DEFINITIONS['line_graph'])
+        peers = graph.peers.values()
+        # use same MPP_SPLIT_PART_FRACTION as in regular LNWallet
+        graph.workers['bob'].MPP_SPLIT_PART_FRACTION = LNWallet.MPP_SPLIT_PART_FRACTION
+
+        # mock split_amount_normal so it's possible to test both cases, the amount getting sorted
+        # out because one part is below the min size and the other case of both parts being just
+        # above the min size, so no part is getting sorted out
+        def mocked_split_amount_normal(total_amount: int, num_parts: int) -> List[int]:
+            if num_parts == 2 and total_amount == 21_000_000:  # test amount 21k sat
+                # this will not get sorted out by suggest_splits
+                return [10_500_000, 10_500_000]
+            elif num_parts == 2 and total_amount == 21_000_001:  # 2nd test case
+                # this will get sorted out by suggest_splits
+                return [11_000_002, 9_999_999]
+            else:
+                return split_amount_normal(total_amount, num_parts)
+
+        async def pay(lnaddr, pay_req):
+            self.assertEqual(PR_UNPAID, graph.workers['alice'].get_payment_status(lnaddr.paymenthash))
+            with mock.patch('electrum.mpp_split.split_amount_normal',
+                                side_effect=mocked_split_amount_normal):
+                result, log = await graph.workers['bob'].pay_invoice(pay_req)
+            self.assertTrue(result)
+            self.assertEqual(PR_PAID, graph.workers['alice'].get_payment_status(lnaddr.paymenthash))
+
+        async def f():
+            async with OldTaskGroup() as group:
+                for peer in peers:
+                    await group.spawn(peer._message_loop())
+                    await group.spawn(peer.htlc_switch())
+                for peer in peers:
+                    await peer.initialized
+                for test in [21_000_000, 21_000_001]:
+                    lnaddr, pay_req = self.prepare_invoice(
+                        graph.workers['alice'],
+                        amount_msat=test,
+                        include_routing_hints=True,
+                        invoice_features=LnFeatures.BASIC_MPP_OPT
+                                         | LnFeatures.PAYMENT_SECRET_REQ
+                                         | LnFeatures.VAR_ONION_REQ
+                    )
+                    await pay(lnaddr, pay_req)
+                raise PaymentDone()
+        with self.assertRaises(PaymentDone):
+            await f()
+
     async def test_payment_multihop(self):
         graph = self.prepare_chans_and_peers_in_graph(self.GRAPH_DEFINITIONS['square_graph'])
         peers = graph.peers.values()
@@ -1727,6 +1865,7 @@ class TestPeerForwarding(TestPeer):
         ):
             alice_w = graph.workers['alice']
             bob_w = graph.workers['bob']
+            carol_w = graph.workers['carol']
             dave_w = graph.workers['dave']
             if mpp_invoice:
                 dave_w.features |= LnFeatures.BASIC_MPP_OPT
@@ -1748,6 +1887,16 @@ class TestPeerForwarding(TestPeer):
                 await asyncio.sleep(2)
             if result:
                 self.assertEqual(PR_PAID, dave_w.get_payment_status(lnaddr.paymenthash))
+                # check mpp is cleaned up
+                async with OldTaskGroup() as g:
+                    for peer in peers:
+                        await g.spawn(peer.wait_one_htlc_switch_iteration())
+                # wait another iteration
+                async with OldTaskGroup() as g:
+                    for peer in peers:
+                        await g.spawn(peer.wait_one_htlc_switch_iteration())
+                for peer in peers:
+                    self.assertEqual(len(peer.lnworker.received_mpp_htlcs), 0)
                 raise PaymentDone()
             elif len(log) == 1 and log[0].failure_msg.code == OnionFailureCode.MPP_TIMEOUT:
                 raise PaymentTimeout()
@@ -1851,46 +2000,52 @@ class TestPeerForwarding(TestPeer):
             await f()
 
     async def _run_trampoline_payment(
-            self, graph, *,
+            self, graph: Graph, *,
             include_routing_hints=True,
             test_hold_invoice=False,
             test_failure=False,
-            attempts=2):
+            attempts=2,
+            sender_name="alice",
+            destination_name="dave",
+            tf_names=("bob", "carol"),
+    ):
 
-        bob_w = graph.workers['bob']
-        carol_w = graph.workers['carol']
-        dave_w = graph.workers['dave']
+        sender_w = graph.workers[sender_name]
+        dest_w = graph.workers[destination_name]
 
         async def pay(lnaddr, pay_req):
-            self.assertEqual(PR_UNPAID, graph.workers['dave'].get_payment_status(lnaddr.paymenthash))
-            result, log = await graph.workers['alice'].pay_invoice(pay_req, attempts=attempts)
+            self.assertEqual(PR_UNPAID, dest_w.get_payment_status(lnaddr.paymenthash))
+            result, log = await sender_w.pay_invoice(pay_req, attempts=attempts)
+            async with OldTaskGroup() as g:
+                for peer in peers:
+                    await g.spawn(peer.wait_one_htlc_switch_iteration())
+            for peer in peers:
+                self.assertEqual(len(peer.lnworker.active_forwardings), 0)
             if result:
-                self.assertEqual(PR_PAID, graph.workers['dave'].get_payment_status(lnaddr.paymenthash))
-                self.assertFalse(bool(bob_w.active_forwardings))
-                self.assertFalse(bool(carol_w.active_forwardings))
+                self.assertEqual(PR_PAID, dest_w.get_payment_status(lnaddr.paymenthash))
                 raise PaymentDone()
             else:
                 raise NoPathFound()
 
         async def f():
-            await self._activate_trampoline(graph.workers['alice'])
+            await self._activate_trampoline(sender_w)
             async with OldTaskGroup() as group:
                 for peer in peers:
                     await group.spawn(peer._message_loop())
                     await group.spawn(peer.htlc_switch())
                 for peer in peers:
                     await peer.initialized
-                lnaddr, pay_req = self.prepare_invoice(dave_w, include_routing_hints=include_routing_hints)
-                self.prepare_recipient(dave_w, lnaddr.paymenthash, test_hold_invoice, test_failure)
+                lnaddr, pay_req = self.prepare_invoice(dest_w, include_routing_hints=include_routing_hints)
+                self.prepare_recipient(dest_w, lnaddr.paymenthash, test_hold_invoice, test_failure)
                 await group.spawn(pay(lnaddr, pay_req))
 
         peers = graph.peers.values()
 
         # declare routing nodes as trampoline nodes
-        electrum.trampoline._TRAMPOLINE_NODES_UNITTESTS = {
-            graph.workers['bob'].name: LNPeerAddr(host="127.0.0.1", port=9735, pubkey=graph.workers['bob'].node_keypair.pubkey),
-            graph.workers['carol'].name: LNPeerAddr(host="127.0.0.1", port=9735, pubkey=graph.workers['carol'].node_keypair.pubkey),
-        }
+        electrum.trampoline._TRAMPOLINE_NODES_UNITTESTS = {}
+        for tf_name in tf_names:
+            peer_addr = LNPeerAddr(host="127.0.0.1", port=9735, pubkey=graph.workers[tf_name].node_keypair.pubkey)
+            electrum.trampoline._TRAMPOLINE_NODES_UNITTESTS[graph.workers[tf_name].name] = peer_addr
 
         await f()
 
@@ -1909,7 +2064,8 @@ class TestPeerForwarding(TestPeer):
         graph = self.prepare_chans_and_peers_in_graph(graph_definition)
         if test_mpp_consolidation:
             graph.workers['dave'].features |= LnFeatures.BASIC_MPP_OPT
-            graph.workers['alice'].network.config.TEST_FORCE_MPP = True
+            graph.workers['alice'].network.config.TEST_FORCE_MPP = True # trampoline must wait until all incoming htlcs are received before sending outgoing htlcs
+            graph.workers['bob'].network.config.TEST_FORCE_MPP = True   # trampoline must wait until all outgoing htlcs have failed before failing incoming htlcs
         if is_legacy:
             # turn off trampoline features in invoice
             graph.workers['dave'].features = graph.workers['dave'].features ^ LnFeatures.OPTION_TRAMPOLINE_ROUTING_OPT_ELECTRUM
@@ -1931,6 +2087,7 @@ class TestPeerForwarding(TestPeer):
             await self._run_trampoline_payment(graph, test_hold_invoice=True, test_failure=True)
 
     async def test_payment_trampoline_legacy(self):
+        # alice -> T1_bob -> carol -> dave
         with self.assertRaises(PaymentDone):
             graph = self.create_square_graph(direct=False, is_legacy=True)
             await self._run_trampoline_payment(graph, include_routing_hints=True)
@@ -1938,13 +2095,34 @@ class TestPeerForwarding(TestPeer):
             graph = self.create_square_graph(direct=False, is_legacy=True)
             await self._run_trampoline_payment(graph, include_routing_hints=False)
 
-    async def test_payment_trampoline_e2e_direct(self):
+    async def test_payment_trampoline_e2e_alice_t1_dave(self):
         with self.assertRaises(PaymentDone):
             graph = self.create_square_graph(direct=True, is_legacy=False)
             await self._run_trampoline_payment(graph)
 
-    async def test_payment_trampoline_e2e_indirect(self):
-        # must use two trampolines
+    async def test_payment_trampoline_e2e_alice_t1_t2_dave(self):
         with self.assertRaises(PaymentDone):
             graph = self.create_square_graph(direct=False, is_legacy=False)
             await self._run_trampoline_payment(graph)
+
+    async def test_payment_trampoline_e2e_alice_t1_carol_t2_edward(self):
+        # alice -> T1_bob -> carol -> T2_dave -> edward
+        graph_definition = self.GRAPH_DEFINITIONS['line_graph']
+        graph = self.prepare_chans_and_peers_in_graph(graph_definition)
+        inject_chan_into_gossipdb(
+            channel_db=graph.workers['bob'].channel_db, graph=graph,
+            node1name='carol', node2name='dave')
+        with self.assertRaises(PaymentDone):
+            await self._run_trampoline_payment(
+                graph, sender_name='alice', destination_name='edward',tf_names=('bob', 'dave'))
+
+
+class TestPeerDirectAnchors(TestPeerDirect):
+    TEST_ANCHOR_CHANNELS = True
+
+class TestPeerForwardingAnchors(TestPeerForwarding):
+    TEST_ANCHOR_CHANNELS = True
+
+
+def run(coro):
+    return asyncio.run_coroutine_threadsafe(coro, loop=util.get_asyncio_loop()).result()
